# standard imports
from argparse import ArgumentParser
import pathlib as pl
from typing import List
import json
import getpass
import datetime

# external imports
import typer # using for quick build of cli prototype

# project imports
from wigeon.packages import Package
from wigeon.db import Connector, Migration

#################################
## Module level variables
#################################
app = typer.Typer()
user = getpass.getuser()

#################################
## Module level functions
#################################



#################################
## CLI Commands
#################################
@app.command()
def createpackage(
    packagename: str,
    dbtype: str,
    environments: str="local,dev,qa,prod"
):
    """
    createpackage initializes a package of migrations in the current
    environment. A package is linked directly to a database type and the
    deployment environments in your ci/cd pipeline.
    """
    typer.echo(f"Creating {packagename} package")
    typer.echo(f"{packagename}'s Database type is {dbtype}")
    typer.echo(f"{packagename}'s environments include {environments.split(',')}")
     # check if package exists
    package = Package(packagename=packagename)
    package.exists(
        raise_error_on_exists=True,
        raise_error_on_not_exist=False
    )

    # initialize package folder
    package.create(
        env_list=environments.split(","),
        db_engine=dbtype
    )


@app.command()
def createmigration(
    migrationname: str,
    packagename: str,
    build: str
):
    """
    createmigration initializes a .sql module for a migration
    """
    typer.echo(f"Creating {migrationname} in {packagename} package...")
    # check if package exists
    package = Package(packagename=packagename)
    package.exists()
    # find latest migration number
    current_migrations = package.list_migrations()
    current_migr_num = package.find_current_migration(migration_list=current_migrations)
    typer.echo(f"Current migration is: {current_migr_num}")
    # create migration
    package.add_migration(
        current_migration=current_migr_num,
        migration_name=migrationname,
        builds=[build] # TODO enable multiple build support at later date
    )
    typer.echo(f"Successfully created {current_migr_num}-{migrationname}.sql!!")

@app.command()
def listmigrations(
    packagename: str
):
    """
    listmigrations lists out all migrations for a given package name
    """
    # check if package exists
    package = Package(packagename=packagename)
    package.exists()
    typer.echo(f"Found following migrations for {packagename}:")
    current_migrations = package.list_migrations()
    for m in sorted(current_migrations):
        typer.echo(f"    {m.name}")
    current_migr = package.find_current_migration(migration_list=current_migrations)
    typer.echo(f"Current migration would be: {current_migr}")

@app.command()
def connect(
    packagename: str,
    server: str=None,
    database: str=None,
    username: str=None,
    password: str=None,
    driver: str=None,
    connectionstring: str=None,
    environment: str=None
):
    """
    connects to a database
    """
    # check if package exists
    package = Package(packagename=packagename)
    package.exists()
    package.read_manifest()
    # create connection, return cursor
    cnctr = Connector(
        package=package,
        environment=environment
    )

    cnxn = cnctr.connect(
        server=server,
        database=database,
        username=username,
        password=password,
        driver=driver,
        connectionstring=connectionstring
    )
    typer.echo(f"Successfully connected to {package.manifest['db_engine']} database!!!!")

@app.command()
def runmigrations(
    packagename: str,
    server: str=None, # connection variable
    database: str=None, # connection variable
    username: str=None, # connection variable
    password: str=None, # connection variable
    driver: str=None, # connection variable
    connectionstring: str=None, # connection variable
    environment: str=None, # migration manifest variable
    all: bool=False, # migration manifest variable
    buildtag: str=None # migration manifest variable
):
    """
    connects to a database and runs migrations
    """
    # check if package exists and read manifest
    package = Package(packagename=packagename)
    package.exists()
    package.read_manifest()
    # create connection

    cnctr = Connector(
        package=package,
        environment=environment
    )

    cnxn = cnctr.connect(
        server=server,
        database=database,
        username=username,
        password=password,
        driver=driver,
        connectionstring=connectionstring
    )
    cur = cnxn.cursor()
    typer.echo(f"Successfully connected to {package.manifest['db_engine']} database!!!!")

    # initialize changelog table if not exists and add columns
    # change_id, migration_date, applied_by(username), and migration_name(.sql filename)
    query_create_changelog = """
    CREATE TABLE IF NOT EXISTS changelog (change_id INTEGER NOT NULL PRIMARY KEY, migration_date TEXT, migration_name TEXT, applied_by TEXT);
    """
    cur.execute(query_create_changelog)
    cnxn.commit()
    
    # find migrations already in target database
    query_migrations_from_changelog = """
    SELECT migration_name from changelog
    """
    cur.execute(query_migrations_from_changelog)
    db_migrations = [n[0] for n in cur.fetchall()]

    # find migrations in manifest
    # filter to migrations only with certain build tag
    mani_migrations = [Migration(**m) for m in package.fetch_manifest_migrations(buildtag=buildtag)]

    # find migrations alead in the database
    duplicate_migrations = [m.name for m in mani_migrations if m.name in db_migrations]
    print(f"Migrations already in db: {duplicate_migrations}")
    # remove duplicate migrations from manifest, unless all option is given
    if not all:
        mani_migrations = [m for m in mani_migrations if m.name not in db_migrations]
    print(f"Migrations to run: {mani_migrations}")
    for mig in mani_migrations:
        if mig.name in db_migrations:
            duplicate_migrations.append(mig.name)
            continue
        mig.run(
            package=package,
            cursor=cur,
            user=user
        )
    print(f"Successfully ran {len(mani_migrations)} migrations")
    cnxn.commit()
    cnxn.close()

if __name__ == "__main__":
    app()


# standard imports
import pathlib as pl
import shutil
import json
from typing import List

# external imports

# project imports


class Package(object):
    pack_folder = pl.Path().cwd().joinpath("packages")

    def __init__(
        self,
        packagename:str
    ):
        self.packagename = packagename
        self.pack_path = Package.pack_folder.joinpath(packagename)
        self.manifest = None

    def exists(
        self,
        raise_error_on_exists: bool=False,
        raise_error_on_not_exist: bool=True
    ) -> bool:
        """
        Checks if a given package exists. Defaults to error if package is not present.
        """
        package_exists = Package.pack_folder.joinpath(self.packagename).exists()
        if package_exists and raise_error_on_exists:
            raise FileExistsError(
                f"A package with the name {self.packagename} already exists at {Package.pack_folder.joinpath(self.packagename)}."
            )
        elif (not package_exists) and (raise_error_on_not_exist):
            raise FileExistsError(
                f"A package with the name {self.packagename} does not exist at {Package.pack_folder.joinpath(self.packagename)}."
            )
        return package_exists
    
    def create(
        self,
        env_list: list,
        db_engine: str
    ):
        """
        Package.create initializes a package in the root/packages/ directory and
        supplies an __init__.py and manifest.json
        """
        # initialize package folder
        self.pack_path.mkdir(parents=True)
        with open(self.pack_path.joinpath("__init__.py"), "w") as f:
            f.write("# auto generated package initializer")
        
        # initialize package manifest
        manifest_template = {}
        manifest_template["db_engine"] = db_engine
        manifest_template["environments"] = {}
        for e in env_list:
            manifest_template["environments"][e] = {
                "connectionstring": None,
                "server": None,
                "database": None,
                "username": None,
                "password:": None
            }
        manifest_template["migrations"] = []

        with open(self.pack_path.joinpath("manifest.json"), "w") as f:
            json.dump(manifest_template, f, indent=4)
    
    def delete(self):
        """
        Package.delete removes the package directory from os
        """
        shutil.rmtree(self.pack_path)
    
    def list_migrations(
        self
    ) -> list:
        """
        list_mgrations reads a package and returns a list of all the migrations
        """
        return [f for f in Package.pack_folder.joinpath(self.packagename).iterdir() if f.suffix == ".sql"]

    def find_current_migration(
        self,
        migration_list: List[pl.Path]
    ) -> str:
        """
        Parses migration files and finds version number to return the latest as a string.
        
        Assumes migration filename convention of:
        ####-<migration_name>.sql
        """
        # BUG latest migration algorithm is erroring sometimes
        if migration_list == []:
            latest_migration_int = 0
        else:
            # sort migration list to ensure largest number is the latest
            migration_list = sorted(migration_list)
            latest_migration_int = int(
                str(
                    migration_list[-1].name
                )[0:4]
            )
        
        # increment migration guid
        current_migration = latest_migration_int + 1
        # check if migration greater than limit of 4 characters
        if current_migration > 9999:
            raise ValueError(
                f"Current migration number ({current_migration}) cannot exceed 9999, max characters allowed in migration guid is 4"
            )
        # return 4 character migration guid string left padded with zeros
        return str(current_migration).zfill(4)

    def read_manifest(self):
        with open(self.pack_path.joinpath("manifest.json"), "r") as f:
            self.manifest = json.load(f)

    def write_manifest(self):
        if not self.manifest:
            raise ValueError(
                f"{self.__class__} {self.packagename}'s manifest not yet read or built. Wigeon will not write nonetype manifest."
            )
        with open(self.pack_path.joinpath("manifest.json"), "w") as f:
            json.dump(self.manifest, f, indent=4)
    
    def fetch_manifest_migrations(self, buildtag: str=None):
        """
        fetch_manifest_migrations collects migrations present in the manifest and
        allows for filtering based on buildtag
        """
        if not self.manifest:
            raise ValueError(
                f"{self.__class__} {self.packagename}'s manifest not yet read or built. Wigeon will not write nonetype manifest."
            )
        if buildtag:
            # TODO implement filtering by build tag
            raise NotImplementedError("Build tag filtering not yet implemented!")
            migrations = {k:v for k,v in self.manifest["migrations"] if buildtag in v["builds"]}
        else:
            migrations = self.manifest["migrations"].copy()
        return migrations
    
    def add_migration(
        self,
        current_migration: str,
        migration_name: str,
        builds: List[str]
    ):
        with open(self.pack_path.joinpath(f"{current_migration}-{migration_name}.sql"), "w") as f:
            f.write("-- TODO build migration code")
        self.read_manifest()
        self.manifest["migrations"].append(
            {
                "name": f"{current_migration}-{migration_name}.sql",
                "builds": [b for b in builds]
            }
        )
        self.write_manifest()

class manifest(object):

    def __init__(self):
        pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re

from setuptools import setup


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    path = os.path.join(package, "__init__.py")
    init_py = open(path, "r", encoding="utf8").read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def get_long_description():
    """
    Return the README.
    """
    return open("README.md", "r", encoding="utf8").read()


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

env_marker_below_38 = "python_version < '3.8'"

minimal_requirements = [
    "click>=8.0.3",
    "typer>=0.4.0",
    "pyodbc>=4.0.32"
]


setup(
    name="wigeon",
    version=get_version("wigeon"),
    url="https://github.com/JLRitch/wigeon",
    license="BSD",
    description="DB Migrations for the continuous developer.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Jere Ritchie",
    author_email="jereritchie@gmail.com",
    packages=get_packages("wigeon"),
    python_requires=">=3.7",
    install_requires=minimal_requirements,
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython"
    ],
    entry_points="""
    [console_scripts]
    wigeon=wigeon.main:app
    """,
    project_urls={
        "Source": "https://github.com/"
    },
)


# standard imports
import pathlib as pl
import os
import sqlite3
import datetime
from typing import TypedDict, Union, List

# external imports
# import pyodbc

# project imports
from wigeon.packages import Package


class Environment(TypedDict):
    connection_string: str
    server: str
    database: str
    username: str
    password: str

class Connector(object):

    db_engines = [
        "sqlite",
        "mssql",
        "postgres"
    ]

    def __init__(
        self,
        package: Package,
        environment: str=None
    ):
        self.db_engine = package.manifest["db_engine"]
        self.package = package
        self.environment = environment
        self.cnxn = None


    def connect(
        self,
        **kwargs
    ):
        db_engines = {
            "sqlite": self.conn_sqlite,
            "mssql": self.conn_mssql,
            "postgres": self.conn_postgres
        }
       
       # read environment name from Connector and collect envvariable names
       # extract environment variables to kwargs if variables exist
        if self.environment:
            print(f"Connecting to {self.environment} environment...")
            kwargs = self.package.manifest["environments"][self.environment]
            # dictionary comprehension ftwftwftw
            kwargs = {k:os.environ[v] for k,v in kwargs.items() if v}

         # run connection method based on db_engine for package
        db_engines[self.db_engine](**kwargs)
        return self.cnxn
        

    def conn_sqlite(
        self,
        **kwargs
    ) -> sqlite3.Connection:
        """
        Connect to a sqlite database and return conn
        """
        if kwargs["connectionstring"]:
            self.cnxn = sqlite3.connect(kwargs["connectionstring"])
            return self.cnxn

    def conn_mssql(self, **kwargs):
        raise NotImplementedError("conn_mssql is not yet implemented!")
    
    def conn_postgres(self, **kwargs):
        raise NotImplementedError("conn_postgres is not yet implemented!")

class Migration(object):

    def __init__(
        self,
        name: str,
        builds: List[str]
    ):
        self.name = name
        self.builds = builds
    
    def __str__(self):
        return f"name: {self.name}, builds: {self.builds}"
    
    def __repr__(self):
        return f"name: {self.name}, builds: {self.builds}"

    def run(
        self,
        package: Package,
        cursor: Union[sqlite3.Cursor, str], # TODO replace str with pyodbc.cursor once implemented
        user: str
    ):
        with open(package.pack_path.joinpath(self.name), "r") as f:
            query = f.read()
        cursor.execute(query)
        cursor.execute(
            "INSERT INTO changelog (migration_date, migration_name, applied_by) VALUES(:migration_date, :migration_name, :applied_by)",
            {
                "migration_date": datetime.datetime.now().strftime("%Y%m%d-%H%M"),
                "migration_name":self.name,
                "applied_by": user
            }
        )